class Colors:
    """ANSI color codes"""

    BLACK = "\033[0;30m"
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    BROWN = "\033[0;33m"
    BLUE = "\033[0;34m"
    PURPLE = "\033[0;35m"
    CYAN = "\033[0;36m"
    LIGHT_GRAY = "\033[0;37m"
    DARK_GRAY = "\033[1;30m"
    LIGHT_RED = "\033[1;31m"
    LIGHT_GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    LIGHT_BLUE = "\033[1;34m"
    LIGHT_PURPLE = "\033[1;35m"
    LIGHT_CYAN = "\033[1;36m"
    LIGHT_WHITE = "\033[1;37m"
    BOLD = "\033[1m"
    FAINT = "\033[2m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    NEGATIVE = "\033[7m"
    CROSSED = "\033[9m"
    END = "\033[0m"
    # cancel SGR codes if we don't write to a terminal
    if not __import__("sys").stdout.isatty():
        for _ in dir():
            if isinstance(_, str) and _[0] != "_":
                locals()[_] = ""
    else:
        # set Windows console in VT mode
        if __import__("platform").system() == "Windows":
            kernel32 = __import__("ctypes").windll.kernel32
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
            del kernel32


def headline(string):
    print(Colors.NEGATIVE)
    print(" " * (len(string) + 2))
    print(" " + string + " ")
    print(" " * (len(string) + 2))
    print(Colors.END)



import argparse
from glyphsets import build_glyphsapp_filter_list, compare_glyphsets, analyze_font
from fontTools.ttLib import TTFont


def main():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command", required=True)

    # Filter List
    filter_lists_parser = subparsers.add_parser("filter-list", help="Ouput Glyphs.app filter from given glyphset(s).")
    filter_lists_parser.add_argument("glyphsets", nargs="+")
    filter_lists_parser.add_argument("--prod-names", action="store_true", default=False)
    filter_lists_parser.add_argument("-o", "--out", required=True, help="output path")

    # Compare
    compare_parser = subparsers.add_parser(
        "compare",
        help="Compare two or more glyhsets to each other; with later glyphsets being compared to the former.",
    )
    compare_parser.add_argument("glyphsets", nargs="+")

    # Analyze
    analyze_parser = subparsers.add_parser(
        "coverage",
        help="Analyze glyphsets covered by a font binary.",
    )
    analyze_parser.add_argument("font", help="Font binary to analyze")

    # PROCESS
    args = parser.parse_args()

    if args.command == "filter-list":
        build_glyphsapp_filter_list(args.glyphsets, args.out, args.prod_names)

    if args.command == "compare":
        compare_glyphsets(args.glyphsets)

    if args.command == "coverage":
        analyze_font(TTFont(args.font))


if __name__ == "__main__":
    main()



import yaml
import plistlib
import os
import gflanguages
from fontTools.unicodedata.Scripts import NAMES as SCRIPT_NAMES
import unicodedata
import functools
from glyphsLib.glyphdata import get_glyph
from glyphsets.helpers import Colors, headline

try:
    from ._version import version as __version__  # type: ignore
except ImportError:
    __version__ = "0.0.0+unknown"

root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "data"))
tool_folder = os.path.abspath(os.path.dirname(__file__))


def sort_unicodes(a, b):
    glyph_a = get_glyph(a)
    glyph_b = get_glyph(b)
    if glyph_a.unicode and glyph_b.unicode:
        return int(glyph_a.unicode, 16) - int(glyph_b.unicode, 16)
    elif glyph_a.unicode:
        return -1
    elif glyph_b.unicode:
        return 1
    else:
        return 0


def read_nam_file(path):
    character_set = set()
    with open(path, "r") as f:
        nam_lines = f.readlines()
    for line in nam_lines:
        unicode = line.split(" ")[0]
        if unicode.startswith("0x"):
            character_set.add(int(unicode[2:], 16))
    return list(sorted(character_set))


def build_glyphsapp_filter_list(glyphsets, out, use_production_names=False):
    """Build filter lists"""
    glyphsets = sorted(glyphsets)

    if not out.endswith(".plist"):
        raise ValueError("Only .plist files are supported")

    # glyphsapp need a prefix for the out file of "CustomFilter"
    if not os.path.basename(out).startswith("CustomFilter"):
        print("Prefixing 'CustomFilter' to out path since file is intended for Glyphsapp")
        out = os.path.join(os.path.dirname(out), "CustomFilter" + os.path.basename(out))

    # Make .plist
    plist = []

    for glyphset_name in glyphsets:
        plist.append(
            {
                "name": glyphset_name,
                "list": glyphs_in_glyphset(glyphset_name, use_production_names),
            }
        )

    with open(out, "wb") as f:
        plistlib.dump(plist, f)
        print("Wrote", out)


regions = gflanguages.LoadRegions()
languages = gflanguages.LoadLanguages()


def defined_glyphsets():
    definitions_path = os.path.join(os.path.dirname(__file__), "definitions")
    yaml_files = [
        os.path.splitext(f)[0]
        for f in os.listdir(definitions_path)
        if os.path.isfile(os.path.join(definitions_path, f)) and f.endswith(".yaml")
    ]
    return sorted(yaml_files)


def get_script(glyphset_name):
    return glyphset_name.split("_")[1]


def defined_scripts():
    scripts = set()
    for glyphset_name in defined_glyphsets():
        scripts.update([get_script(glyphset_name)])
    return sorted(list(scripts))


def glyphsets_per_script(script):
    glyphsets = []
    for glyphset_name in defined_glyphsets():
        if get_script(glyphset_name) == script:
            glyphsets.append(glyphset_name)
    return glyphsets


def get_glyphset_definition(glyphset_name):
    yaml_path = os.path.join(os.path.dirname(__file__), "definitions", f"{glyphset_name}.yaml")
    with open(yaml_path, "r", encoding="utf8") as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def unicodes_per_glyphset(glyphset_name):
    # Read .nam file
    nam_path = os.path.join(
        os.path.dirname(__file__),
        "results",
        "nam",
        f"{glyphset_name}.nam",
    )
    if os.path.exists(nam_path):
        return read_nam_file(nam_path)


def glyphs_in_glyphsets(glyphset_names, production_names=False):
    glyphs = set()
    for glyphset_name in glyphset_names:
        glyphs.update(glyphs_in_glyphset(glyphset_name, production_names))

    return sorted(list(glyphs))


def glyphs_in_glyphset(glyphset_name, production_names=False):
    # script = glyphset_name.split("_")[1]

    with open(
        os.path.join(
            tool_folder,
            "results",
            "txt",
            "nice-names",
            f"{glyphset_name}.txt",
        ),
        "r",
    ) as f:
        glyph_names = [line.strip() for line in f.readlines() if not line.startswith("#")]

    return sorted(glyph_names)


def languages_per_glyphset(glyphset_name):

    script = get_script(glyphset_name)
    glyphset_definition = get_glyphset_definition(glyphset_name)
    language_codes = glyphset_definition.get("language_codes", [])
    regions = glyphset_definition.get("regions")
    # use_aux = glyphset_definition.get("use_auxiliary", False)
    historical = glyphset_definition.get("historical", False)
    population = glyphset_definition.get("population", False)
    exclude_language_codes = glyphset_definition.get("exclude_language_codes", set())

    # Assemble character sets from gflanguages
    if regions:
        for language in languages.values():
            if language.id in exclude_language_codes:
                continue
            if not historical and language.historical:
                continue
            if population > language.population:
                continue
            if set(language.region).intersection(set(regions)) and SCRIPT_NAMES[language.script] == script:
                language_codes.append(language.id)

    return language_codes


def categorize_glyphs(glyph_names):
    categories = {}

    unicode_sorted_glyph_names = sorted(glyph_names, key=functools.cmp_to_key(sort_unicodes))

    for glyph_name in unicode_sorted_glyph_names:
        glyph = get_glyph(glyph_name)
        if not glyph.category:
            category = "Uncategorized"
        else:
            category = glyph.category
            if category == "Mark" and glyph.subCategory == "Nonspacing":
                category = "Mark, nonspacing"
            elif category == "Mark":
                category = "Mark, spacing"
        if category not in categories:
            categories[category] = []

        if glyph.unicode:
            categories[category].append(chr(int(glyph.unicode, 16)))
        else:
            categories[category].append("/" + glyph_name)

    return dict(sorted(categories.items()))


def add_dotted_circle(character):
    return "◌" + character


def describe_glyphset(glyph_names, target="markdown", color=""):
    md = ""
    categories = categorize_glyphs(glyph_names)
    for category, characters in categories.items():
        md += f"{Colors.BOLD if target=='console' else ''}{category}{Colors.END if target=='console' else ''} "
        md += f"({len(characters)} glyphs): \n"

        if category == "Mark, nonspacing":
            string = " ".join(map(add_dotted_circle, characters))
        else:
            string = " ".join(characters)
        if target == "markdown":
            string = string.replace("`", "/grave")
        md += (
            f"`{color if target=='console' and color else ''}"
            + string
            + f"{Colors.END if target=='console' and color else ''}`\n\n"
        )
    return md


def compare_glyphsets(glyphsets):
    """
    Compare the contents of glyphsets to each other,
    with the first glyphset being the reference.
    Each consecutive glyphset gets compares to each former.
    """

    if len(list(set(glyphsets))) != len(glyphsets):
        raise ValueError("Please provide unique glyphsets to compare.")

    reference_glyphset = glyphsets[0]
    reference_glyphs = set(glyphs_in_glyphset(reference_glyphset))

    if not reference_glyphs:
        raise ValueError(f"Glyphset {reference_glyphset} is empty.")

    headline(reference_glyphset)
    print(f"Total glyphs: {len(reference_glyphs)}\n")
    print(describe_glyphset(reference_glyphs, target="console"))

    for i, _glyphset in enumerate(glyphsets):
        if i == 0:
            continue

        this_glyphs = set(glyphs_in_glyphset(glyphsets[i]))
        if not this_glyphs:
            raise ValueError(f"Glyphset {glyphsets[i]} is empty.")

        if i > 1:
            previous_glyphs = set(glyphs_in_glyphset(glyphsets[i - 1]))
        else:
            previous_glyphs = reference_glyphs

        headline(glyphsets[i])
        print(f"Total glyphs: {len(this_glyphs)}\n")

        missing = previous_glyphs.difference(this_glyphs)
        extra = this_glyphs.difference(previous_glyphs)

        if extra:
            print(
                f"{Colors.BOLD}{glyphsets[i]}{Colors.END} has {len(extra)} {Colors.GREEN}{Colors.BOLD}additional"
                + f"{Colors.END} glyphs compared to {Colors.BOLD}{glyphsets[i - 1]}{Colors.END}:\n"
            )
            print(describe_glyphset(extra, target="console", color=Colors.GREEN))

        if missing:
            print(
                f"{Colors.BOLD}{glyphsets[i]}{Colors.END} is {Colors.RED}{Colors.BOLD}missing{Colors.END} "
                + f"{len(missing)} glyphs compared to {Colors.BOLD}{glyphsets[i - 1]}{Colors.END}:\n"
            )
            print(describe_glyphset(missing, target="console", color=Colors.RED))


def add_country(code):
    if code in regions:
        return f"{regions[code].name} ({code})"
    return code


def add_language(code):
    if code in languages:
        return f"{languages[code].name} ({code})"
    return code


def description_per_glyphset(glyphset_name):
    script = get_script(glyphset_name)
    glyphset_definition = get_glyphset_definition(glyphset_name)
    language_codes = glyphset_definition.get("language_codes", [])
    regions = glyphset_definition.get("regions")
    use_aux = glyphset_definition.get("use_auxiliary", False)
    historical = glyphset_definition.get("historical", False)
    population = glyphset_definition.get("population", False)
    description = glyphset_definition.get("description", None)
    exclude_language_codes = glyphset_definition.get("exclude_language_codes", [])

    glyphs_stub_path = os.path.join(root_folder, "definitions", "per_glyphset", f"{glyphset_name}.stub.glyphs")

    warning = False
    md = ""

    md += f"# {glyphset_name.replace('_', ' ')}\n\n"
    if description:
        md += (
            "> _Description partially salvaged from old README, so languages manually listed here (if any) may be "
            + "outdated or irrelevant and need to be replaced by language code lists:_\n> \n> "
            + "\n> ".join(description.split("\n"))
            + "\n\n"
        )
    if regions:
        md += f"`{glyphset_name}` is **dynamically** defined [here](/Lib/glyphsets/definitions/{glyphset_name}.yaml) "
        md += "as:\n\n"
    else:
        md += f"`{glyphset_name}` is **statically** defined [here](/Lib/glyphsets/definitions/{glyphset_name}.yaml) "
        md += "as:\n\n"
    md += f"* Script: {script}\n"

    # Dynamic defintion
    if regions:
        md += "* All languages of the countries `\n" + ",\n".join(sorted(map(add_country, regions))) + "\n`\n"
    if population:
        md += f"* With a population of over {population} speakers\n"
    if historical:
        md += "* Including historical languages\n"
    if use_aux:
        md += "* Including auxiliary characters\n"
    if exclude_language_codes:
        md += (
            "* Excluding the following languages: `\n"
            + ",\n".join(sorted(map(add_language, exclude_language_codes)))
            + "\n`\n"
        )

    if regions and language_codes:
        md += (
            "* Additionally, the following languages are defined **statically**: `\n"
            + ",\n".join(sorted(map(add_language, language_codes)))
            + "\n`\n"
        )

    # Static defintion
    elif not regions and language_codes:
        md += "* List of languages: `\n" + ",\n".join(sorted(map(add_language, language_codes))) + "\n`\n"

    # Additional resources
    if os.path.exists(glyphs_stub_path):
        md += f"* Characters and glyphs defined in [{os.path.basename(glyphs_stub_path)}]"
        md += f"(/data/definitions/per_glyphset/{os.path.basename(glyphs_stub_path)})\n"
    for language_code in language_codes:
        lang_stub_path = os.path.join(root_folder, "definitions", "per_language", f"{language_code}.stub.glyphs")
        if os.path.exists(lang_stub_path):
            md += "* Language-specific characters and glyphs defined for "
            md += (
                f"[{add_language(language_code)}](/data/definitions/per_language/{os.path.basename(lang_stub_path)})\n"
            )

    if not regions and not language_codes:
        md += "\n> [!CAUTION]  \n> Since this glyphset has no defined languages, it can't be checked via "
        md += "Fontbakery's `shape_languages` check.\n> Please add language code definions "
        md += f"[here](/Lib/glyphsets/definitions/{glyphset_name}.yaml).\n"
        warning = True

    md += "\n"

    if regions:
        _languages_per_glyphset = languages_per_glyphset(glyphset_name)
        md += (
            f"\nThe following list of **{len(_languages_per_glyphset)}** languages is computed as a result "
            + "of the dynamic conditions described above:\n\n`\n"
            + ",\n".join(sorted(map(add_language, _languages_per_glyphset)))
            + "\n`\n\n"
        )

    # Content
    md += "### Characters and Glyphs\n\n"
    md += str(describe_glyphset(glyphs_in_glyphset(glyphset_name)))

    # Composed characters
    decomposed_chars = get_decomposed_chars(glyphset_name)
    if decomposed_chars:
        md += "### Character Sequences\n\n"
        md += f"The following {len(decomposed_chars)} composed character sequences are decomposed in the font:\n\n"
        md += "`\n"
        md += " ".join(decomposed_chars)
        md += "\n`\n\n"

    md += "### Resulting Glyphset Files\n\n"
    md += f".nam file (only encoded characters): [{glyphset_name}.nam](/data/results/nam/{glyphset_name}.nam)\n\n"
    md += f"Glyphs.app source file: [{glyphset_name}.glyphs](/data/results/glyphs/{glyphset_name}.glyphs)\n\n"
    md += f"Text files: [{glyphset_name}.txt](/data/results/txt/nice-names/{glyphset_name}.txt) (nice names) and "
    md += f"[{glyphset_name}.txt](/data/results/txt/prod-names/{glyphset_name}.txt) (production names)\n\n"
    md += f"Glyphs.app Custom Filter List (contains all {script} glyphsets): "
    md += f"[CustomFilter_GF_{script}.plist](/data/results/plist/CustomFilter_GF_{script}.plist)\n\n"

    return md, warning


def get_glyphsets_fulfilled(ttFont):
    """Returns a dictionary of glyphsets that are fulfilled by the font,
    and the percentage of glyphs in the font that are in the glyphset.
    This is following the new glyphset definitions in glyphsets
    """

    res = {}
    unicodes_in_font = set(ttFont.getBestCmap().keys())
    for glyphset in defined_glyphsets():
        unicodes_in_glyphset = unicodes_per_glyphset(glyphset)
        if glyphset not in res:
            res[glyphset] = {"has": [], "missing": [], "percentage": 0}
        for unicode in unicodes_in_glyphset:
            if unicode in unicodes_in_font:
                res[glyphset]["has"].append(unicode)
            else:
                res[glyphset]["missing"].append(unicode)

        if len(res[glyphset]["has"]) > 0 and len(res[glyphset]["missing"]) == 0:
            res[glyphset]["percentage"] = 1
        else:

            # Calculate the percentage supported not in total glyphs in font
            # compared to the total glyphs in the glyphset,
            # but shrink both to the unique set of Unicodes in the glyphset.
            # Compare everything to Latin Core:
            # - Subtract Latin Core from the covered Unicodes
            # - Calculate percentage covered based on that

            if glyphset == "GF_Latin_Core":
                if unicodes_in_glyphset:
                    has = res[glyphset]["has"]
                    res[glyphset]["percentage"] = len(has) / len(unicodes_in_glyphset)
                else:
                    res[glyphset]["percentage"] = 0
            else:
                unicodes_unique_in_glyphset = list(
                    set(unicodes_per_glyphset(glyphset)).difference(set(unicodes_per_glyphset("GF_Latin_Core")))
                )

                if unicodes_unique_in_glyphset:

                    has = set(res[glyphset]["has"]).intersection(set(unicodes_unique_in_glyphset))
                    res[glyphset]["percentage"] = len(has) / len(unicodes_unique_in_glyphset)

                    # res[glyphset]["has_unique"] = list(has)
                    # res[glyphset]["unique_in"] = unicodes_unique_in_glyphset
                else:
                    res[glyphset]["percentage"] = 0
    return dict(sorted(res.items(), key=lambda x: x[1]["percentage"], reverse=True))


def get_decomposed_chars(glyphset_name):
    allchars = set()
    for lang_code in languages_per_glyphset(glyphset_name):
        lang = languages[lang_code]
        allchars.update(getattr(getattr(lang, "exemplar_chars", {}), "base", "").split(" "))
        if get_glyphset_definition(glyphset_name).get("use_auxiliary"):
            allchars.update(getattr(getattr(lang, "exemplar_chars", {}), "auxiliary", "").split(" "))
    decomposed_chars = sorted(
        [
            g[1:-1]
            for g in allchars
            if any(unicodedata.combining(c) for c in g) and g.startswith("{") and g.endswith("}")
        ]
    )

    return decomposed_chars


def analyze_font(ttFont):
    results = get_glyphsets_fulfilled(ttFont)

    def print_support(lower=0, upper=1):

        if lower >= 0.8:
            print(
                (
                    f"{Colors.BROWN}These glyphsets will {Colors.ITALIC}implicitly{Colors.END}{Colors.BROWN} be part "
                    f"of Fontbakery's {Colors.ITALIC}shape_languages{Colors.END}{Colors.BROWN} check:{Colors.END}"
                )
            )
            print()

        found = 0
        for key in results:
            if (lower < upper and lower <= results[key]["percentage"] < upper) or (
                lower == upper and results[key]["percentage"] == lower
            ):

                color = color = Colors.RED if results[key]["percentage"] < 0.8 else Colors.GREEN

                found += 1

                not_covered = ""
                if lower >= 0.8:
                    languages = languages_per_glyphset(key)
                    if not languages:
                        not_covered = (
                            f"{Colors.BROWN}(Not part of {Colors.ITALIC}shape_languages{Colors.END}"
                            f"{Colors.BROWN} because of missing language definitions){Colors.END}"
                        )
                print(
                    (
                        f"{Colors.BOLD}{key}{Colors.END} {color}{int(results[key]['percentage']*100)}%"
                        f"{Colors.END} {not_covered}"
                    )
                )
                if results[key]["missing"]:
                    print(f"Missing: {' '.join([chr(x) for x in results[key]['missing']])}")
                    if "unique" in results[key]:
                        print(f"Unique: {' '.join([chr(x) for x in results[key]['unique']])}")
                    if "has_unique" in results[key]:
                        print(f"Unique in font: {' '.join([chr(x) for x in results[key]['has_unique']])}")
                    print()
        if not found:
            print("———")

    headline("Fully supported glyphsets:")
    print_support(1, 1)

    print()
    headline("Partially supported glyphsets (>80%):")
    print_support(0.8, 1)

    print()
    headline("Unsupported glyphsets (<80%):")
    print_support(0, 0.8)


if __name__ == "__main__":
    print(defined_scripts())



"""
Assemble .nam files from .nam stub files and language definitions.
"""

import sys
import os
import shutil
import gflanguages
import unicodedata
import glyphsLib
import functools
import plistlib
from glyphsLib.glyphdata import get_glyph, _lookup_attributes_by_unicode
from fontTools.unicodedata.Scripts import NAMES as SCRIPT_NAMES

# Insert local module path at beginning of sys.path
# so that up-to-date version of glyphsets package is used
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "Lib"))
from glyphsets import (
    get_script,
    defined_glyphsets,
    get_glyphset_definition,
    unicodes_per_glyphset,
    languages_per_glyphset,
    read_nam_file,
)  # noqa: E402


def sort_unicodes(a, b):
    if a.unicode and b.unicode:
        return int(a.unicode, 16) - int(b.unicode, 16)
    elif a.unicode:
        return -1
    elif b.unicode:
        return 1
    else:
        return 0


def sort_by_category(a, b):
    info_a = get_glyph(a.name)
    info_b = get_glyph(b.name)

    if info_a.category is None:
        return -1
    elif info_b.category is None:
        return 1

    value = sorted([info_a.category, info_b.category]).index(info_a.category)
    if value == 0:
        value = -1
    value *= -1

    return value


def assemble_characterset(root_folder, glyphset_name):

    script = get_script(glyphset_name)
    glyphset_definition = get_glyphset_definition(glyphset_name)
    language_codes = languages_per_glyphset(glyphset_name)
    use_aux = glyphset_definition.get("use_auxiliary", False)

    nam_stub_path = os.path.join(
        root_folder, "definitions", "per_glyphset", f"{glyphset_name}.stub.nam"
    )
    nam_path = os.path.join(root_folder, "results", "nam", f"{glyphset_name}.nam")
    nam_in_package_path = os.path.abspath(
        os.path.join(
            root_folder,
            "..",
            "Lib",
            "glyphsets",
            "results",
            "nam",
            f"{glyphset_name}.nam",
        )
    )
    txt_in_package_path = os.path.abspath(
        os.path.join(
            root_folder,
            "..",
            "Lib",
            "glyphsets",
            "results",
            "txt",
            "nice-names",
            f"{glyphset_name}.txt",
        )
    )
    glyphs_stub_path = os.path.join(
        root_folder, "definitions", "per_glyphset", f"{glyphset_name}.stub.glyphs"
    )
    glyphs_path = os.path.join(
        root_folder, "results", "glyphs", f"{glyphset_name}.glyphs"
    )
    glyphs_empty_path = os.path.join(root_folder, f"empty_font.glyphs")
    txt_nicenames_path = os.path.join(
        root_folder, "results", "txt", "nice-names", f"{glyphset_name}.txt"
    )
    txt_prodnames_path = os.path.join(
        root_folder, "results", "txt", "prod-names", f"{glyphset_name}.txt"
    )
    plist_path = os.path.join(
        root_folder, "results", "plist", f"CustomFilter_GF_{script}.plist"
    )

    character_set = set()

    # Assemble character sets from gflanguages
    languages = gflanguages.LoadLanguages()
    for language_code in language_codes:
        chars = languages[language_code].exemplar_chars
        # chars.base.upper() is important because many Latin languages don't
        # contain a complete set of uppercase letters in "index"
        character_set.update(
            {
                ord(c)
                for c in list(
                    set(chars.base)
                    | set(chars.base.upper())
                    | set(chars.index)
                    | set(chars.marks)
                    | set(chars.numerals)
                    | set(chars.punctuation)
                    | (set(chars.auxiliary) if use_aux else set())
                )
                if c not in (" ", "{", "}", "◌")
            }
        )

    # Call get_glyph once so that GLYPHDATA gets filled in glyphsLib
    get_glyph("A")
    # If I import GLYPHDATA at the top of the file, it doesn't get filled
    from glyphsLib.glyphdata import GLYPHDATA

    assert type(GLYPHDATA) is glyphsLib.glyphdata.GlyphData

    def _font_has_unicode(font, unicode):
        for glyph in font.glyphs:
            if glyph.unicode:
                if int(glyph.unicode, 16) == unicode:
                    return True

    # Create or open glyphs file and add characters
    if os.path.exists(glyphs_stub_path):
        font = glyphsLib.load(glyphs_stub_path)
        for glyph in font.glyphs:
            if glyph.unicodes:
                for unicode in glyph.unicodes:
                    character_set.update({int(unicode, 16)})
    else:
        font = glyphsLib.load(glyphs_empty_path)

    font.familyName = glyphset_name

    # Add language-specific glyphs
    for language_code in language_codes:
        per_language_glyphs_stub_path = os.path.join(
            root_folder, "definitions", "per_language", f"{language_code}.stub.glyphs"
        )
        if os.path.exists(per_language_glyphs_stub_path):
            per_language_font = glyphsLib.load(per_language_glyphs_stub_path)

            for glyph in per_language_font.glyphs:

                # Add encoded characters to character_set
                if glyph.unicodes:
                    for unicode in glyph.unicodes:
                        character_set.update({int(unicode, 16)})

                # Add unencoded glyphs to .glyphs file
                else:
                    new_glyph = glyphsLib.GSGlyph(glyph.name)
                    font.glyphs.append(new_glyph)

    # Add encoded characters to .glyphs file
    for _i, unicode in enumerate(sorted(list(character_set))):
        if not _font_has_unicode(font, unicode):
            unicode = f"{unicode:#0{6}X}".replace("0X", "")
            glyph_info = _lookup_attributes_by_unicode(unicode, GLYPHDATA)
            if "name" in glyph_info:
                new_glyph = glyphsLib.GSGlyph(glyph_info["name"])
            else:
                new_glyph = glyphsLib.GSGlyph(f"uni{unicode}")
            new_glyph.unicode = unicode
            font.glyphs.append(new_glyph)

    # Sort
    font.glyphs = sorted(font.glyphs, key=functools.cmp_to_key(sort_by_category))
    unicode_sorted_glyphs = sorted(font.glyphs, key=functools.cmp_to_key(sort_unicodes))
    glyph_names = [glyph.name for glyph in unicode_sorted_glyphs]
    production_glyph_names = [
        get_glyph(glyph.name).production_name for glyph in unicode_sorted_glyphs
    ]

    # Save glyphs file
    os.makedirs(os.path.dirname(glyphs_path), exist_ok=True)
    font.save(glyphs_path)

    # Output sorted character set to .nam file
    os.makedirs(os.path.dirname(nam_path), exist_ok=True)
    os.makedirs(os.path.dirname(nam_in_package_path), exist_ok=True)
    with open(nam_path, "w") as f:
        f.write(
            "# This file is auto-generated; do not edit. See /README.md for instructions.\n"
        )
        for i, unicode in enumerate(sorted(list(character_set))):
            unicode_string = f"{unicode:#0{6}X}".replace("0X", "0x")
            try:
                unicode_name = unicodedata.name(chr(unicode))
            except ValueError:
                unicode_name = ""
            f.write(f"{unicode_string} {unicode_name}")
            if i < len(character_set) - 1:
                f.write("\n")
    shutil.copyfile(nam_path, nam_in_package_path)

    # Output txt files
    os.makedirs(os.path.dirname(txt_nicenames_path), exist_ok=True)
    with open(txt_nicenames_path, "w") as f:
        f.write(
            "# This file is auto-generated; do not edit. See /README.md for instructions.\n"
        )
        f.write("\n".join(glyph_names))
    os.makedirs(os.path.dirname(txt_prodnames_path), exist_ok=True)
    with open(txt_prodnames_path, "w") as f:
        f.write(
            "# This file is auto-generated; do not edit. See /README.md for instructions.\n"
        )
        f.write("\n".join(production_glyph_names))
    os.makedirs(os.path.dirname(txt_in_package_path), exist_ok=True)
    shutil.copyfile(txt_nicenames_path, txt_in_package_path)

    # Adjust .plist
    os.makedirs(os.path.dirname(plist_path), exist_ok=True)
    if os.path.exists(plist_path):
        with open(plist_path, "rb") as f:
            plist = plistlib.load(f)
    else:
        plist = []
    found_list = False
    for plist_glyphset in plist:
        if "name" in plist_glyphset and plist_glyphset["name"] == glyphset_name:
            plist_glyphset["list"] = glyph_names
            found_list = True
    if not found_list:
        plist.append({"name": glyphset_name, "list": glyph_names})
    with open(plist_path, "wb") as f:
        plistlib.dump(plist, f)


if __name__ == "__main__":
    # Check for gflanguages version
    installed = None
    latest = None
    for line in os.popen("pip index versions gflanguages").read().split("\n"):
        if "INSTALLED" in line:
            installed = line.split(" ")[-1].strip()
        if "LATEST" in line:
            latest = line.split(" ")[-1].strip()
    print(
        f"""
*************************************************************
*
*   WARNING:
*   Make sure you're using the correct version of gflanguages,
*   otherwise the glyphsets will be incorrect.
*
*   You have: {installed}
*   Location: {gflanguages.__file__}
*
*************************************************************
"""
    )

    root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "data"))

    for glyphset_name in defined_glyphsets():
        print(f"Assembling '{glyphset_name}'...")
        assemble_characterset(root_folder, glyphset_name)
        # Proof of work:
        # assert unicodes_per_glyphset(glyphset_name) != []



"""
Browse through all gflanguages and find the input character.

Syntax:
python scripts/find_language_for_character.py <character>
(<character> can be either a single 0x0000 string or unicode character.)

Please manually install tabulate if not present: pip install tabulate
"""

import argparse
import gflanguages
import sys
import tabulate
import unicodedata

regions = gflanguages.LoadRegions()
languages = gflanguages.LoadLanguages()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Find languages using a given character."
    )
    parser.add_argument(
        "character", metavar="CHAR", help="Unicode character or 0x0000 string"
    )
    args = parser.parse_args()

    input_character = args.character
    if input_character.startswith("0x"):
        input_character = chr(int(input_character, 16))
    unicode_string = f"{ord(input_character):#0{6}X}".replace("0X", "0x")
    print(
        f"Character: [{input_character}]  ({unicode_string} {unicodedata.name(input_character)})"
    )

    found_languages = []

    # Read language definitions
    for lang_code in languages:
        lang = languages[lang_code]
        if lang.exemplar_chars:
            chars = lang.exemplar_chars
            for category in ("base", "index", "marks", "numerals", "punctuation"):
                if (
                    input_character in getattr(chars, category)
                    or input_character in getattr(chars, category).upper()
                ):
                    # Find regions
                    found_regions = set()
                    for country_code in regions:
                        if country_code in lang.region:
                            found_regions.update(
                                set(regions[country_code].region_group)
                            )

                    found_languages.append(
                        (
                            lang_code,
                            lang.name,
                            category,
                            lang.population,
                            lang.script,
                            ", ".join(list(found_regions)),
                        )
                    )

    found_languages = sorted(found_languages, key=lambda x: x[3], reverse=True)
    print(
        tabulate.tabulate(
            found_languages,
            headers=["Language", "Name", "Category", "Speakers", "Script", "Regions"],
        )
    )



import os
from glyphsets import (
    unicodes_per_glyphset,
    languages_per_glyphset,
    get_glyphsets_fulfilled,
    defined_glyphsets,
    compare_glyphsets,
    build_glyphsapp_filter_list,
    glyphs_in_glyphset,
)
import plistlib

DATA_FP = os.path.join(os.path.dirname(__file__), "data")
FONT_PATH = os.path.join(DATA_FP, "MavenPro[wght].ttf")


def test_definitions():
    assert len(unicodes_per_glyphset("GF_Latin_Core")) == 319

    assert len(languages_per_glyphset("GF_Arabic_Plus")) == 5
    assert len(languages_per_glyphset("GF_Latin_African")) == 601

    # accidental double definitions
    for code in defined_glyphsets():
        assert len(languages_per_glyphset(code)) == len(set(languages_per_glyphset(code)))
    assert len(defined_glyphsets()) == len(set(defined_glyphsets()))

    assert "GF_Latin_Core" in defined_glyphsets()


def test_coverage():
    from fontTools.ttLib import TTFont

    ttFont = TTFont(FONT_PATH)
    assert get_glyphsets_fulfilled(ttFont)["GF_Latin_Core"]["percentage"] > 0.99


def test_compare():
    compare_glyphsets(["GF_Latin_Kernel", "GF_Latin_Core", "GF_Latin_Plus"])


def test_filter_lists():
    build_glyphsapp_filter_list(["GF_Latin_Kernel", "GF_Latin_Core", "GF_Latin_Plus"], "test.plist", False)
    assert os.path.exists("CustomFiltertest.plist")
    test = plistlib.load(open("CustomFiltertest.plist", "rb"))
    assert len(test) == 3
    assert test[0]["name"] == "GF_Latin_Core"
    assert test[1]["name"] == "GF_Latin_Kernel"
    assert test[2]["name"] == "GF_Latin_Plus"
    assert test[0]["list"] == glyphs_in_glyphset("GF_Latin_Core")
    assert test[1]["list"] == glyphs_in_glyphset("GF_Latin_Kernel")
    assert test[2]["list"] == glyphs_in_glyphset("GF_Latin_Plus")
    os.remove("CustomFiltertest.plist")



"""
Read all languages and regions from gflanguages and print out the languages that are supported
in Europe or Americas and have more than 5 million speakers.
"""

import argparse
import gflanguages
import logging

SPEAKERS = 5000000


def main(args):
    arg_parser = argparse.ArgumentParser(
        description="Generates a list of language ids based on given requirements."
    )
    arg_parser.add_argument(
        "--regions",
        nargs="+",
        help="Region or region groups where languages are used.",
        default=[],
    )
    arg_parser.add_argument(
        "--population", help="Minimum speakers’ population.", default=SPEAKERS
    )
    arg_parser.add_argument(
        "--script",
        nargs="+",
        required=True,
        help="Script of the languages.",
    )
    arg_parser.add_argument(
        "--include-languages",
        nargs="+",
        default=[],
        help="Languages that should be included even if they don’t match the other requirements.",
    )
    arg_parser.add_argument(
        "--exclude-languages",
        nargs="+",
        default=[],
        help="Languages that should be excluded even if they match the other requirements.",
    )
    options = arg_parser.parse_args(args)
    options.population = int(options.population)

    regions = gflanguages.LoadRegions()
    if not options.regions:
        options.regions = [country_code for country_code in regions]
    languages = gflanguages.LoadLanguages()

    country_codes = set()

    # Go through countries and add them to the set if their region_group is in options.regions or they are in options.regions
    for country_code in regions:
        if set(regions[country_code].region_group).intersection(
            set(options.regions)
        ) or (country_code in options.regions):
            country_codes.add(country_code)

    # Go through languages and add them to the set if their regions are in country_codes
    language_codes = set()

    # Print languages if they are supported through the logic
    print("language_codes:")
    for language_code in sorted(languages.keys()):
        if (
            (
                language_code in options.include_languages
                or set(languages[language_code].region).intersection(country_codes)
            ) and language_code not in options.exclude_languages
            and languages[language_code].population >= options.population
            and languages[language_code].script in options.script
        ):
            if not languages[language_code].exemplar_chars:
                logging.warning(
                    f"{languages[language_code].name} has no character definitions."
                )
            language_codes.add(language_code)
            print(f"  - {language_code}  # {languages[language_code].name}")


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



# coding: utf-8
# Copyright 2013 The Font Bakery Authors.
# Copyright 2017-2021 The Google Fonts Tools Authors.
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See AUTHORS.txt for the list of Authors and LICENSE.txt for the License.
import os
from setuptools import setup

# Read the contents of the README file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="glyphsets",
    use_scm_version={"write_to": "Lib/glyphsets/_version.py"},
    url="https://github.com/googlefonts/glyphsets/",
    description="A python API for evaluating coverage of glyph sets in font projects.",
    long_description=long_description,
    long_description_content_type="text/markdown",  # This is important!
    author=(
        "Dave Crossland, "
        "Eli Heuer, "
        "Felipe Sanches, "
        "Lasse Fister, "
        "Marc Foley, "
        "Yanone, "
        "Roderick Sheeter"
    ),
    author_email="dave@lab6.com",
    package_dir={"": "Lib"},
    packages=["glyphsets"],
    package_data={
        "glyphsets": [
            "*.json",
            "definitions/*.yaml",
            "results/nam/*.nam",
            "results/txt/nice-names/*.txt",
        ]
    },
    entry_points={"console_scripts": ["glyphsets = glyphsets.__main__:main"]},
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Text Processing :: Fonts",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.8",
    setup_requires=["setuptools_scm>=8.0.4,<8.1"],
    install_requires=[
        "requests",
        "setuptools",
        "FontTools[ufo]",
        "glyphsLib>=6.7.1",
        "defcon",
        "unicodedata2",
        "gflanguages>=0.6.0",
        "pyyaml",
    ],
    extras_require={
        "dev": [  # For the tests to run
            "gfsubsets",
        ]
    },
)



"""
Assemble .nam files from .nam stub files and language definitions.
"""

import sys
import os

# Insert local module path at beginning of sys.path
# so that up-to-date version of glyphsets package is used
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "Lib"))
from glyphsets import (
    defined_glyphsets,
    defined_scripts,
    glyphsets_per_script,
    description_per_glyphset,
)  # noqa: E402

if __name__ == "__main__":
    root_folder = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "GF_Glyphsets")
    )

    repo_root_folder = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
        )
    )

    md = []

    # TOC
    md.append("# Table of Contents:\n\n")

    # Get scripts
    scripts = defined_scripts()
    script_dict = {}
    max_glyphsets = 0
    md.append("| " + " | ".join(scripts) + " |")
    md.append("| " + " | ".join(["---"] * len(scripts)) + " |")
    for script in scripts:
        max_glyphsets = max(max_glyphsets, len(glyphsets_per_script(script)))
        script_dict[script] = glyphsets_per_script(script)
    for i in range(max_glyphsets):
        row = []
        for script in scripts:
            if i < len(script_dict[script]):
                glyphset_name = script_dict[script][i]
                abbr = glyphset_name.split("_")[-1]
                _new_md, warning = description_per_glyphset(glyphset_name)
                warning_md = "✅"
                if warning:
                    warning_md = "🛑"
                row.append(
                    f"[{warning_md} {abbr}](#{glyphset_name.lower().replace('_', '-')})"
                )
            else:
                row.append("")
        md.append("| " + " | ".join(row) + " |")

    md.append(
        "\n> [!NOTE]  \n> This document is a human-readable representation of the glyphset defintions defined in `.yaml` files [here](/Lib/glyphsets/definitions/) and gets updated automatically as part of the `sh build.sh` command.\n"
    )
    md.append(
        "\n> [!NOTE]  \n> The symbols ✅ and 🛑 above denote whether or not this glyphset is available as part of Fontbakery's `shape_languages` check; in other words, whether or not language codes are defined for it.\n"
    )

    md.append("\n")

    # Content
    for glyphset_name in defined_glyphsets():
        new_md, _warning = description_per_glyphset(glyphset_name)
        md.append(new_md)

    with open(os.path.join(repo_root_folder, "GLYPHSETS.md"), "w") as f:
        f.write("\n".join(md))
